/* AMX Mod script.
 *
 * (c) 2003, OLO
 * This file is provided as is (no warranties).
 *
 * Cvars:
 * amx_reserved_slots < value > - sets how many slots we want to keep for admins
 * amx_reservation < value > - sets kicking rules:
 * 0 - Disables checking for slots usage
 * 1 - Kick by shortes time a player without reservation when server goes full
 *     until there are so many players with reservation as amx_reserved_slots specifies.
 * 2 - Kick by worst ping a player without reservation when server goes full
 *     until there are so many players with reservation as amx_reserved_slots specifies.
 * 3 - Kick a player without reservation who tries to enter to one of the reservered slots
 *     which amount is set by amx_reserved_slots.
 */

#include <amxmod>
#include <amxmisc>

// Comment if you don't want to ban ppl who are
// keep reconnecting more than three times in a very
// short time even if there is no free slots
//#define BAN_RECONNECTING

// Comment if you don't want to hide not used reserved slots
// NOTE: Status of free slots is changed at someone
// connection and disconnection
#define HIDE_RESERVEDSLOTS

// Uncomment if you are using Steam (Counter-Strine 1.6)
//#define STEAM_VERSION

#if defined BAN_RECONNECTING
new g_lastCon[32]
new g_lastConNum
new Float:g_lastCon_time
new g_logBan[] = "Reservation: ^"%s<%d><%s><>^" ban reconnecter for 1 minute"
#endif

new g_logFile[16]
new g_logDrop[] = "Reservation: ^"%s<%d><%s><>^" drop due to slot reservation"
new g_reservedLeftCvar[] =  "amx_real_res_slots" // "amx_reservedleft"
new g_visibleMaxCvar[] = "sv_visiblemaxplayers"

public plugin_init()
{
  register_plugin("Slots Reservation","0.9","default")
  register_cvar("amx_reserved_slots","2")
  register_cvar(g_reservedLeftCvar,"0",FCVAR_SERVER|FCVAR_UNLOGGED)
  register_cvar("amx_reservation","3")
  get_logfile(g_logFile,15)
}

kickPlayer(id)
{
  // gather some info
  new name[32], authid[32]
  get_user_name(id,name,31)
  get_user_authid(id,authid,31)
  
#if defined BAN_RECONNECTING  
  // save who and how often was connecting
  new Float:gameTime = get_gametime()
  if ( !equal(g_lastCon,name) || g_lastCon_time + 5.0 < gameTime ) 
  {
    g_lastCon = name
    g_lastConNum = 0
  }
  else
    ++g_lastConNum
  g_lastCon_time = gameTime
#endif
  
  // if bot then kick with console cmd
  if ( is_user_bot(id) ) 
  {
    server_cmd("kick #%d", get_user_userid(id) )
    log_to_file(g_logFile,g_logDrop,name, get_user_userid(id) ,authid)
  }
#if defined BAN_RECONNECTING  
  // reconnecting player - ban him
  else if ( g_lastConNum > 3 ) 
  {
    // leave him information message
    if (g_lastConNum < 5)
    {
      client_cmd(id,"echo ^"You have been banned for 1 minute^"")
      log_to_file(g_logFile,g_logBan,name, get_user_userid(id) ,authid)
    }
    
    server_cmd("banid 1 #%d", get_user_userid(id) )
    client_cmd(id,"disconnect")
  }
#endif   
  // else just disconnect him, leaving a proper message
  else 
  {
#if defined BAN_RECONNECTING
    if ( g_lastConNum > 2 )
      client_cmd(id,"echo ^"Next time you will be banned for 1 minute^"")
#endif      
    client_cmd(id,"echo ^"Dropped due to slot reservation^";disconnect")
    log_to_file(g_logFile,g_logDrop,name, get_user_userid(id) ,authid)
  }
}

kickLag(){
  new who = 0, ping, loss, worst = -1
  new maxplayers = get_maxplayers()
  for(new i = 1; i <= maxplayers; ++i){
    if ( !is_user_connected(i) && !is_user_connecting(i) ) 
      continue // not used slot  
    if (get_user_flags(i)&ADMIN_RESERVATION) 
      continue // has reservation, skip him
    get_user_ping(i,ping,loss) // get ping
    if ( ping > worst ) {
      worst = ping
      who = i
    }
  }
  if (who) kickPlayer(who)
  return who  
}

kickFresh(){
  new who = 0, itime, shortest = 0x7fffffff
  new maxplayers = get_maxplayers()
  for(new i = 1; i <= maxplayers; ++i){
    if ( !is_user_connected(i) && !is_user_connecting(i) )
      continue // not used slot
    if (get_user_flags(i)&ADMIN_RESERVATION)
      continue // has reservation, skip him
    itime = get_user_time(i) // get user playing time with connection duration  
    if ( shortest > itime ) {
      shortest = itime
      who = i
    }
  }
  if (who) kickPlayer(who)
  return who
}

/* If there is at least one place then allow enter the game anyone.
* If the slots are reserved and not all filled then kick
* with amx_reservation rules.
*/

#if defined STEAM_VERSION
public client_putinserver(id)
  set_task(0.5,"client_entered",id)

public client_entered(id){
#else
public client_connect(id){
#endif

  new resType = get_cvar_num( "amx_reservation" ) 
  if ( !resType ) {
    set_cvar_num( g_reservedLeftCvar , 0 )
    set_cvar_num( g_visibleMaxCvar , -1 )
    return PLUGIN_CONTINUE
  }
  new reservedMax, reservedNum, playersMax, playersNum, kicked
  getSlotUsage(playersNum,playersMax,reservedNum,reservedMax)
  if ( (kicked = freeSlot( get_user_flags(id)&ADMIN_RESERVATION , 
    playersNum,playersMax,reservedNum,reservedMax,resType)) == -1 ) {
    kickPlayer(id)
    return PLUGIN_HANDLED // block connect in other plugins (especially in consgreet)
  }
  else if ( kicked ){
    new name[32]
    get_user_name( kicked, name , 31 )
    client_cmd(id,"echo ^"* %s was kicked to free this slot^"" ,name )
  }
  setSlotUsage( resType , playersNum , playersMax ,  reservedNum , reservedMax)
  return PLUGIN_CONTINUE
}

public client_disconnect(id){
  new resType = get_cvar_num( "amx_reservation" ) 
  if ( !resType ) {
    set_cvar_num( g_reservedLeftCvar , 0 )
    set_cvar_num( g_visibleMaxCvar , -1 )
    return PLUGIN_CONTINUE
  }
  new reservedMax, reservedNum, playersNum, playersMax
  getSlotUsage(playersNum,playersMax,reservedNum,reservedMax)
  setSlotUsage( resType , --playersNum , playersMax , 
    (get_user_flags(id)&ADMIN_RESERVATION) ?  reservedNum - 1 : reservedNum , reservedMax)
  return PLUGIN_CONTINUE
}

setSlotUsage( resType, playersNum , playersMax ,  reservedNum , reservedMax  ) {

  if ( resType < 3 ) {
    if ( reservedMax <= reservedNum  ) {
      set_cvar_num( g_reservedLeftCvar , 0  )
#if defined HIDE_RESERVEDSLOTS      
      set_cvar_num( g_visibleMaxCvar , playersMax )
#endif
    }
    else {
      set_cvar_num( g_reservedLeftCvar , reservedMax - reservedNum  )
#if defined HIDE_RESERVEDSLOTS
      set_cvar_num( g_visibleMaxCvar , playersMax - 1 )
#endif
    }
  }
  else if (  playersNum > playersMax - reservedMax ){
    set_cvar_num( g_reservedLeftCvar , playersMax - playersNum )
#if defined HIDE_RESERVEDSLOTS
    set_cvar_num( g_visibleMaxCvar , playersNum )
#endif
  }
  else {
    set_cvar_num( g_reservedLeftCvar , reservedMax  )
#if defined HIDE_RESERVEDSLOTS
    set_cvar_num( g_visibleMaxCvar , playersMax - reservedMax )
#endif
  }
}

getSlotUsage(&pNum,&pMax,&rNum,&rMax)
{
  pMax = get_maxplayers()
  rMax = get_cvar_num("amx_reserved_slots")  
  pNum = rNum = 0
  for(new i = 1; i <= pMax; ++i) {
    if ( !is_user_connected(i) && !is_user_connecting(i) ) 
      continue // not used slot
    if (get_user_flags(i)&ADMIN_RESERVATION)
      ++rNum
    ++pNum
  }
}

freeSlot( hasRes ,playersNum,playersMax,reservedNum,reservedMax,resType)
{
  if ( resType < 3 ) { // This is for amx_reservation 1 and 2
    if (playersNum < playersMax) // Don't care until all slots are filled (f.e. 12/12)
      return 0
    if ( hasRes ) { // Connecting player is an admin
      if ( reservedMax > reservedNum ) { // And there are still free reserved slots
        switch(resType) {     // so kick one players to make server from 12/12 to 11/12
          case 1: return kickFresh()
          case 2: return kickLag()
        }
      }
    }
    else if ( reservedMax > reservedNum ) 
      return -1 // This is not an admin and there are still free reserved slots
  } 
  // This is for amx_reservation 3
  else if ( !hasRes && playersNum > playersMax - reservedMax )
    return -1
  return 0
}
